! Dot product test for linear Radon operator
program Dot_product_test_lwivu

  use sep
  use mem_mod
  use fbi_mod
  use hess_wemva_mod
  use solver_mod
!  use lradon_mod
  use dptest

  implicit none

  integer                             :: nt, n1, n2, nlag1, nlag2, wfld_j, i, gpu_n, gpu_o, niter
  integer                             :: sou_nx, sou_ox, sou_dx, sou_z, sou_x, isou
  integer                             :: rec_nx, rec_ox, rec_dx, rec_z, rec_x, irec
  integer                             :: bc1, bc2, bc3, bc4, ntau, otau
  real                                :: dt, d1, o1, d2, o2
  real                                :: lambda
  integer                             :: n_model, n_data, stat

  logical                             :: dipole
 
  real, dimension(:), allocatable     :: wavelet
  real, dimension(:,:), allocatable   :: vel
  real, dimension(:,:,:), allocatable :: data_obs
    real, dimension(:), allocatable   :: mod1, mod2
    real, dimension(:), allocatable   :: dat1, dat2
!  integer            :: nz, np, nt, nx, n_model, n_data
!  real               :: oz, dz, op, dp, ot, dt, ox, dx
  real, dimension(2) :: dot1, dot2

  call sep_init()

  call from_history("n1", n1)
  call from_history("o1", o1)
  call from_history("d1", d1)
  call from_history("n2", n2)
  call from_history("o2", o2)
  call from_history("d2", d2)

  call from_param("ntau", ntau, 1)
  call from_param("otau", otau, 0)
  if(ntau<1-otau) write(0,*) "WARNING: zero subsurface axis is not included"

  call from_param("nt", nt)
  call from_aux("wavelet", "d1", dt)

  call from_param("bc1", bc1)
  call from_param("bc2", bc2)
  call from_param("bc3", bc3)
  call from_param("bc4", bc4)
  call from_param("wfld_j", wfld_j)

  call from_param("sou_nx", sou_nx)
  call from_param("sou_ox", sou_ox)
  call from_param("sou_dx", sou_dx)
  call from_param("sou_z", sou_z)
  if(sou_ox < 1 .or. (sou_nx-1)*sou_dx+sou_ox > n2) call erexit("sou axis values are not correct")
  if(sou_z < 1 .or. sou_z > n1) call erexit("sou_z value is not correct")

  call from_param("rec_nx", rec_nx)
  call from_param("rec_ox", rec_ox)
  call from_param("rec_dx", rec_dx)
  call from_param("rec_z", rec_z)
  if(rec_ox < 1 .or. (rec_nx-1)*rec_dx+rec_ox > n2) call erexit("rec axis values are not correct")
  if(rec_z < 1 .or. rec_z > n1) call erexit("rec_z value is not correct")

  call from_param("dipole", dipole, .false.)

!  if(oper > 0) call from_param("adj", adj)
  call from_param("gpu_n", gpu_n)
  call from_param("gpu_o", gpu_o)

  call from_param("lambda",lambda,0.1)

  call from_param("nlag1", nlag1, 0)
  call from_param("nlag2", nlag2, 0)

  call init1(n1, o1, d1, n2, o2, d2, bc1, bc2, bc3, bc4)
  call init2(sou_nx, sou_ox, sou_dx, sou_z, rec_nx, rec_ox, rec_dx, rec_z)
  call init3(nt, dt, wfld_j, ntau, otau, gpu_n, gpu_o)
  call init4(dipole)

!  call from_param("nz", nz)
!  call from_param("oz", oz)
!  call from_param("dz", dz)
!  call from_param("np", np)
!  call from_param("op", op)
!  call from_param("dp", dp)
!  call from_param("nt", nt)
!  call from_param("ot", ot)
!  call from_param("dt", dt)
!  call from_param("nx", nx)
!  call from_param("ox", ox)
!  call from_param("dx", dx)

  call mem_alloc1d(wavelet, nt)
  call mem_alloc2d(vel, n1, n2)
  call mem_alloc3d(data_obs, nt, rec_nx, sou_nx)

  call sep_read(vel)
  call sep_read(wavelet, "wavelet")

  call init_alloc(wavelet)
  call model_update(vel)

  write(0,*) "Using dipole source? ",dipole
  write(0,*) " "
  write(0,*) "Reading data..."
  do i=1,sou_nx
    write(0,*) "Reading source: ",i,"of ",sou_nx
    call sep_read(data_obs(:,:,i), "data")
  end do
  call data_update(data_obs)

  n_model = 2*n1*n2
  n_data = 2*n1*n2*ntau

  call H_init(n1, n2, nlag1, nlag2)
  call hess_wemva_init(n_data, n_model, lambda, .true.) 

  call mem_alloc1d(mod1, n_model)
  call mem_alloc1d(mod2, n_model)
  call mem_alloc1d(dat1, n_data)
  call mem_alloc1d(dat2, n_data)
  call dot_test(hess_wemva, n_model, n_data, dot1, dot2)

  call to_history("dot1", dot1)
  call to_history("dot2", dot2)
 
  write(0,*) dot1
  write(0,*) dot2


!  call sep_close()

end program
