program oper3

  use sep
  use mem_mod
  use omp_lib
  use fbi_energ_mod

  implicit none

  integer                             :: stat
  integer                             :: nt, n1, n2, wfld_j, i, gpu_n, gpu_o
  integer                             :: sou_nx, sou_ox, sou_dx, sou_z, sou_x, isou
  integer                             :: rec_nx, rec_ox, rec_dx, rec_z, rec_x, irec
  integer                             :: bc1, bc2, bc3, bc4, ntau, otau
  real                                :: dt, d1, o1, d2, o2
  real                                :: alpha

  integer                             :: oper
  logical                             :: adj, dipole, verbose

  real, dimension(:), allocatable     :: wavelet, data, model
  real, dimension(:,:), allocatable   :: vel, deltaB
  real, dimension(:,:,:), allocatable :: refl, data_obs, deltaD, deltaI
  real, dimension(:,:,:,:), allocatable :: reflenergy

  integer, dimension(8)               :: timer0, timer1
  real                                :: timer_sum=0.

  call sep_init()

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer0)
!---------------------------------------------------------------------------

  call from_param("oper", oper)
  if(oper < 0 .or. oper > 3) call erexit("oper value is not correct")

  call from_history("n1", n1)
  call from_history("o1", o1)
  call from_history("d1", d1)
  call from_history("n2", n2)
  call from_history("o2", o2)
  call from_history("d2", d2)

  call from_param("ntau", ntau, 1)
  call from_param("otau", otau, 0)
  if(ntau<1-otau) write(0,*) "WARNING: zero subsurface axis is not included"

  call from_param("nt", nt)
  call from_aux("wavelet", "d1", dt)

  call from_param("bc1", bc1)
  call from_param("bc2", bc2)
  call from_param("bc3", bc3)
  call from_param("bc4", bc4)
  call from_param("wfld_j", wfld_j)

  call from_param("sou_nx", sou_nx)
  call from_param("sou_ox", sou_ox)
  call from_param("sou_dx", sou_dx)
  call from_param("sou_z", sou_z)
  if(sou_ox < 1 .or. (sou_nx-1)*sou_dx+sou_ox > n2) call erexit("sou axis values are not correct")
  if(sou_z < 1 .or. sou_z > n1) call erexit("sou_z value is not correct")

  call from_param("rec_nx", rec_nx)
  call from_param("rec_ox", rec_ox)
  call from_param("rec_dx", rec_dx)
  call from_param("rec_z", rec_z)
  if(rec_ox < 1 .or. (rec_nx-1)*rec_dx+rec_ox > n2) call erexit("rec axis values are not correct")
  if(rec_z < 1 .or. rec_z > n1) call erexit("rec_z value is not correct")

  call from_param("dipole", dipole, .false.)
  call from_param("verbose", verbose, .true.)
  call from_param("alpha", alpha, 0.)

  if(oper > 0) call from_param("adj", adj)
  call from_param("gpu_n", gpu_n)
  call from_param("gpu_o", gpu_o)

  call init1(n1, o1, d1, n2, o2, d2, bc1, bc2, bc3, bc4)
  call init2(sou_nx, sou_ox, sou_dx, sou_z, rec_nx, rec_ox, rec_dx, rec_z)
  call init3(nt, dt, wfld_j, ntau, otau, gpu_n, gpu_o)
  call init4(dipole, verbose, alpha)


  if((oper .eq. 0) .or. (oper < 3 .and. .not. adj)) then
    call to_history("n1", nt)
    call to_history("o1", 0.)
    call to_history("d1", dt)
    call to_history("n2", rec_nx)
    call to_history("o2", rec_ox)
    call to_history("d2", rec_dx)
    call to_history("n3", sou_nx)
    call to_history("o3", sou_ox)
    call to_history("d3", sou_dx)
  else
    call to_history("n1", n1)
    call to_history("o1", o1)
    call to_history("d1", d1)
    call to_history("n2", n2)
    call to_history("o2", o2)
    call to_history("d2", d2)
    if((oper .eq. 1))then
      call to_history("n3", ntau)
      call to_history("o3", otau)
      call to_history("d3", 1.)
      call to_history("n1", n1,   "ei90")
      call to_history("o1", o1,   "ei90")
      call to_history("d1", d1,   "ei90")
      call to_history("n2", n2,   "ei90")
      call to_history("o2", o2,   "ei90")
      call to_history("d2", d2,   "ei90")
      call to_history("n3", ntau, "ei90")
      call to_history("o3", otau, "ei90")
      call to_history("d3", 1.,   "ei90")
      call to_history("n1", n1,   "cic")
      call to_history("o1", o1,   "cic")
      call to_history("d1", d1,   "cic")
      call to_history("n2", n2,   "cic")
      call to_history("o2", o2,   "cic")
      call to_history("d2", d2,   "cic")
      call to_history("n3", ntau, "cic")
      call to_history("o3", otau, "cic")
      call to_history("d3", 1.,   "cic")
    elseif(oper .eq. 3 .and. .not. adj)then
      call to_history("n3", ntau)
      call to_history("o3", otau)
      call to_history("d3", 1.)
    end if
  end if

  call mem_alloc1d(wavelet, nt)
  call mem_alloc2d(vel, n1, n2)

  call mem_alloc1d(model, n1*n2)
  call mem_alloc1d(data, n1*n2*ntau)

  call sep_read(vel)
  call sep_read(wavelet, "wavelet")

  call init_alloc(wavelet)
  call model_update(vel)

  if(oper .eq. 0) then
    call mem_alloc3d(data_obs, nt, rec_nx, sou_nx)
    call L_nlop2(.false., data_obs)
    write(0,*) "Nonlinear WE operator running!"
    write(0,*) "Using dipole source? ",dipole
    write(0,*) " "
    do i=1,sou_nx
      write(0,*) "Source No. ",i
      call sep_write(data_obs(:,:,i))
    end do
  else if(oper .eq. 1) then
!    call omp_set_num_threads(4)
    call mem_alloc4d(reflenergy, n1, n2, ntau, 3)
    call mem_alloc3d(data_obs, nt, rec_nx, sou_nx)
    if(adj) then
      do i=1,sou_nx
        call sep_read(data_obs(:,:,i), "data")
      end do
      call L_lop2(adj, .false., reflenergy, data_obs)
      do i=1,ntau
        call sep_write(reflenergy(:,:,i,1))
      end do
      do i=1,ntau
        call sep_write(reflenergy(:,:,i,2),"ei90")
      end do
      do i=1,ntau
        call sep_write(reflenergy(:,:,i,3),"cic")
      end do
    else
      do i=1,ntau
        call sep_read(reflenergy(:,:,i,1), "refl")
      end do
      call L_lop2(adj, .false., reflenergy(:,:,:,1), data_obs)
      do i=1,sou_nx
        call sep_write(data_obs(:,:,i))
      end do
    end if
  else if(oper .eq. 2) then
    call mem_alloc3d(refl, n1, n2, ntau)
    call mem_alloc2d(deltaB, n1, n2)
    call mem_alloc3d(deltaD, nt, rec_nx, sou_nx)
    do i=1,ntau
      call sep_read(refl(:,:,i), "refl")
    end do
    call refl_update(refl)
    if(adj) then
      do i=1,sou_nx
        call sep_read(deltaD(:,:,i), "deltaD")
      end do
      call T_lop2(adj, .false., deltaB, deltaD)
      call sep_write(deltaB)
    else
      call sep_read(deltaB, "deltaB")
      call T_lop2(adj, .false., deltaB, deltaD)
      do i=1,sou_nx
        call sep_write(deltaD(:,:,i))
      end do
    end if
  else if(oper .eq. 3) then
    call mem_alloc3d(data_obs, nt, rec_nx, sou_nx)
    call mem_alloc2d(deltaB, n1, n2)
    call mem_alloc3d(deltaI, n1, n2, ntau)
    write(0,*) "Using dipole source? ",dipole
    write(0,*) " "
    write(0,*) "Reading data..."
    do i=1,sou_nx
      write(0,*) "Reading source: ",i
      call sep_read(data_obs(:,:,i), "data")
    end do
    call data_update(data_obs)
    write(0,*) "Finish reading data..."
    if(adj) then
      write(0,*) "Adjoint WEMVA operator running!"
      do i=1,ntau
        call sep_read(deltaI(:,:,i), "deltaI")
      end do
      data = reshape(deltaI,(/n1*n2*ntau/))
!      call W_lop2(adj, .false., deltaB, deltaI)
      stat=W_lop(adj, .false., model, data)
      write(0,*) "Finish adjoint WEMVA operator, now writing!"
      deltaB = reshape(model,(/n1,n2/))
      call sep_write(deltaB)
    else
      write(0,*) "Forward WEMVA operator running!"
      call sep_read(deltaB, "deltaB")
!      call W_lop2(adj, .false., deltaB, deltaI)
      model = reshape(deltaB,(/n1*n2/))
      stat=W_lop(adj, .false., model, data)
      deltaI = reshape(data,(/n1,n2,ntau/))
      write(0,*) "Finish forward WEMVA operator, now writing!"
      do i=1,ntau
        call sep_write(deltaI(:,:,i))
      end do
    end if
  end if

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  write(0,*) "DONE!"
!---------------------------------------------------------------------------

end program
