module energyr0_mod
  
  use gradient_mod
  implicit none

  integer, private                             :: n1, n2, nt, ntau, otau, it, itau, tau, i1, i2
  real, private                                :: cos2th, pi
  real, dimension(:,:,:), allocatable, private :: StRt
  real, dimension(:,:,:,:), pointer, private   :: squareR

  contains

  subroutine energyr0_init(n1_in, n2_in, nt_in, ntau_in, otau_in, squareR_in)
    integer                         :: n1_in, n2_in, nt_in, ntau_in, otau_in
    real, dimension(:,:,:,:), target :: squareR_in
    n1 = n1_in
    n2 = n2_in
    nt = nt_in
    ntau = ntau_in
    otau = otau_in
    squareR => squareR_in
    allocate(StRt(n1,n2,nt))
    pi = 3.14159265358979323846
  end subroutine

  function energyr0_op(adj, add, model, data, angrej) result(stat)
    logical,intent(in) :: adj, add
    real,dimension(:)  :: model, data
    integer            :: stat, angrej
    call energyr0_op2(adj, add, model, data, angrej)
    stat=0
  end function

  subroutine energyr0_op2(adj, add, model, data, angrej, St, Rt)
    optional                    :: St, Rt
    integer, intent(in)         :: angrej
    logical,intent(in)          :: adj, add 
    real, dimension(n1,n2,ntau) :: model
    real, dimension(n1,n2,nt,3) :: data
    real, dimension(n1,n2,nt)   :: St, Rt
    cos2th = cos(2.*angrej*pi/180.)
    StRt = 0
    if(present(St)) StRt = cos2th*St*Rt
    if(adj) then
      if(.not. add) model = 0.
      do itau=1,ntau
        tau = itau - 1 + otau
        do it=1+abs(tau),nt-abs(tau)
          do i2 = 1,n2
            do i1 = 1,n1
              model(i1,i2,itau) = model(i1,i2,itau) + &
                                 (data(i1,i2,it+tau,3)*squareR(i1,i2,it-tau,3)*cos2th+  & 
                                  data(i1,i2,it+tau,1)*squareR(i1,i2,it-tau,1)+ &
                                  data(i1,i2,it+tau,2)*squareR(i1,i2,it-tau,2)+ &
                                  StRt(i1,i2,it))
            end do
          end do
        end do
      end do
    else
      if(.not. add) model = 0.
      do itau=1,ntau
        tau = itau - 1 + otau
        do it=1+abs(tau),nt-abs(tau)
          do i2 = 1,n2
            do i1 = 1,n1
              model(i1,i2,itau) = model(i1,i2,itau) - &
                                 (data(i1,i2,it+tau,3)*squareR(i1,i2,it-tau,3)*cos2th+  & 
                                  data(i1,i2,it+tau,1)*squareR(i1,i2,it-tau,1)+ &
                                  data(i1,i2,it+tau,2)*squareR(i1,i2,it-tau,2)+ &
                                  StRt(i1,i2,it))
            end do
          end do
        end do
      end do
    end if
  end subroutine

end module
