module energys0_mod

  use gradient_mod
  implicit none

  integer, private                              :: n1, n2, nt, ntau, otau
  integer, private                              :: it, itau, tau, i1, i2
  real, private                                 :: cos2th, pi
  real, dimension(:,:,:), allocatable, private  :: temp1, temp2
  real, dimension(:,:,:,:), pointer, private    :: squareS

  contains

  subroutine energys0_init(n1_in, n2_in, nt_in, ntau_in, otau_in, squareS_in)
    integer                         :: n1_in, n2_in, nt_in, ntau_in, otau_in
    real, dimension(:,:,:,:), target :: squareS_in
    n1 = n1_in
    n2 = n2_in
    nt = nt_in
    ntau = ntau_in
    otau = otau_in
    squareS => squareS_in
    allocate(temp1(n1,n2,ntau))
    allocate(temp2(n1,n2,ntau))
    pi = 3.14159265358979323846
  end subroutine

  function energys0_op(adj, add, model, data, angrej) result(stat)
    logical,intent(in) :: adj, add
    real,dimension(:)  :: model, data
    integer            :: stat, angrej
    call energys0_op2(adj, add, model, data, angrej)
    stat=0
  end function

  subroutine energys0_op2(adj, add, model, data, angrej)
    integer, intent(in)         :: angrej
    logical,intent(in)          :: adj, add 
    real, dimension(n1,n2,ntau) :: model
    real, dimension(n1,n2,nt,3) :: data
    cos2th = cos(2.*angrej*pi/180.)
    if(adj) then
      if(.not. add) model = 0.
      do itau=1,ntau
        tau = itau - 1 + otau
        do it=1+abs(tau),nt-abs(tau)
          do i2 = 1,n2
            do i1 = 1,n1
              model(i1,i2,itau) = model(i1,i2,itau) + &
                                 (data(i1,i2,it+tau,3)*squareS(i1,i2,it-tau,3)*cos2th+  & 
                                  data(i1,i2,it+tau,1)*squareS(i1,i2,it-tau,1)+ &
                                  data(i1,i2,it+tau,2)*squareS(i1,i2,it-tau,2))
            end do
          end do
        end do
      end do
    else
      if(.not. add) model = 0.
      do itau=1,ntau
        tau = itau - 1 + otau
        do it=1+abs(tau),nt-abs(tau)
          do i2 = 1,n2
            do i1 = 1,n1
              model(i1,i2,itau) = model(i1,i2,itau) - &
                                 (data(i1,i2,it+tau,3)*squareS(i1,i2,it-tau,3)*cos2th+  & 
                                  data(i1,i2,it+tau,1)*squareS(i1,i2,it-tau,1)+ &
                                  data(i1,i2,it+tau,2)*squareS(i1,i2,it-tau,2))
            end do
          end do
        end do
      end do
    end if
  end subroutine

  subroutine eic_th0(alpha, model90, model0)
    real, intent(in)            :: alpha
    real, dimension(n1,n2,ntau) :: model0
    real, dimension(n1,n2,ntau) :: model90

    temp1 = abs(model0)+abs(model90)
    where(temp1 == 0.)
      temp2 = 0.
    elsewhere
      temp2 = abs(model90)/temp1
    end where
    model0 = model0*exp(-alpha*temp2)
  end subroutine
end module
