#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cudaErrors.cu"

#define dloc(z,x) ((z)+(x)*n1_gpu)

#define BLOCK_SIZE 16
#define FAT 4
#define COEFFS_SIZE 8 
#define CZ1 0
#define CX1 1
#define CZ2 2
#define CX2 3
#define CZ3 4
#define CX3 5
#define CZ4 6
#define CX4 7

__constant__ int n1_gpu, n2_gpu, nt_gpu;
__constant__ float *p1_gpu, *p1x_gpu, *p1z_gpu;
__constant__ float coeffs[COEFFS_SIZE];

__global__ void grad_for_kernel(float *p1, float *p1x, float *p1z){
  __shared__ float temp[BLOCK_SIZE+2*FAT][BLOCK_SIZE+2*FAT];
  register int ig=blockIdx.x*blockDim.x+threadIdx.x;   // global z coordinates
  register int jg=blockIdx.y*blockDim.y+threadIdx.y;   // global x coordinates
  register int il=FAT+threadIdx.x;               // thread-local z coordinate
  register int jl=FAT+threadIdx.y;               // thread-local x coordinate
  register int addr=dloc(ig,jg);                  // linear memroy address

  temp[jl][il] = (ig<n1_gpu && jg<n2_gpu) ? p1[dloc(ig,jg)] : 0.0;
  if(threadIdx.x<FAT) temp[jl][il-FAT] = (ig-FAT>=0 && ig-FAT<n1_gpu && jg<n2_gpu) ? p1[dloc(ig-FAT,jg)] : 0.0;
  if(threadIdx.x>=BLOCK_SIZE-FAT) temp[jl][il+FAT] = (ig+FAT<n1_gpu && jg<n2_gpu) ? p1[dloc(ig+FAT,jg)] : 0.0;
  if(threadIdx.y<FAT) temp[jl-FAT][il] = (jg-FAT>=0 && jg-FAT<n2_gpu && ig<n1_gpu) ? p1[dloc(ig,jg-FAT)] : 0.0;
  if(threadIdx.y>=BLOCK_SIZE-FAT) temp[jl+FAT][il] = (jg+FAT<n2_gpu && ig<n1_gpu) ? p1[dloc(ig,jg+FAT)] : 0.0;
/*
  if(threadIdx.x<FAT-1) temp[jl][il-FAT+1] = (ig-FAT+1>=0 && ig-FAT+1<n1_gpu && jg<n2_gpu) ? p1[dloc(ig-FAT+1,jg)] : 0.0;
  if(threadIdx.x>=BLOCK_SIZE-FAT) temp[jl][il+FAT] = (ig+FAT<n1_gpu && jg<n2_gpu) ? p1[dloc(ig+FAT,jg)] : 0.0;
  if(threadIdx.y<FAT-1) temp[jl-FAT+1][il] = (jg-FAT+1>=0 && jg-FAT+1<n2_gpu && ig<n1_gpu) ? p1[dloc(ig,jg-FAT+1)] : 0.0;
  if(threadIdx.y>=BLOCK_SIZE-FAT) temp[jl+FAT][il] = (jg+FAT<n2_gpu && ig<n1_gpu) ? p1[dloc(ig,jg+FAT)] : 0.0;
*/
  __syncthreads();

  if(ig>=n1_gpu || jg>=n2_gpu) return;

  p1x[addr] =
       coeffs[CX4]*(-temp[jl-3][il]+temp[jl+4][il])
      +coeffs[CX3]*(-temp[jl-2][il]+temp[jl+3][il])
      +coeffs[CX2]*(-temp[jl-1][il]+temp[jl+2][il])
      +coeffs[CX1]*(-temp[jl][il]  +temp[jl+1][il]);

  p1z[addr] =
      +coeffs[CZ4]*(-temp[jl][il-3]+temp[jl][il+4])
      +coeffs[CZ3]*(-temp[jl][il-2]+temp[jl][il+3])
      +coeffs[CZ2]*(-temp[jl][il-1]+temp[jl][il+2])
      +coeffs[CZ1]*(-temp[jl][il]  +temp[jl][il+1]);
}
/*
extern "C" void gpu_setdevice(int *gpu_id)
{
  cudaSetDevice(*gpu_id);
}
*/
extern "C" void gpu_grad_init(int *n1_in, int *n2_in, int *nt_in, float *d1_in, float *d2_in)
{
  int n1, n2, nt;
  float d1, d2;
  float *p1, *p1x, *p1z;
  long int memsize;

  n1 = (*n1_in);
  n2 = (*n2_in);
  nt = (*nt_in);
  d1 = (*d1_in);
  d2 = (*d2_in);

  cudaMemcpyToSymbol(n1_gpu,  &n1,  sizeof(int));
  cudaMemcpyToSymbol(n2_gpu,  &n2,  sizeof(int));
  cudaMemcpyToSymbol(nt_gpu,  &nt,  sizeof(int));

  float coeffs_cpu[COEFFS_SIZE];
  coeffs_cpu[CZ1]= (1224.0/1024.0)/d1;
  coeffs_cpu[CX1]= (1224.0/1024.0)/d2;
  coeffs_cpu[CZ2]=-(245.0/3072.0)/d1;
  coeffs_cpu[CX2]=-(245.0/3072.0)/d2;
  coeffs_cpu[CZ3]= (49.0/5120.0)/d1;
  coeffs_cpu[CX3]= (49.0/5120.0)/d2;
  coeffs_cpu[CZ4]=-(5.0/7168.0)/d1;
  coeffs_cpu[CX4]=-(5.0/7168.0)/d2;
  cudaMemcpyToSymbol(coeffs, coeffs_cpu, COEFFS_SIZE*sizeof(float));

  cudaMalloc( (void**) &p1,       n1*n2*sizeof(float));
  cudaMalloc( (void**) &p1x,      n1*n2*sizeof(float));
  cudaMalloc( (void**) &p1z,      n1*n2*sizeof(float));


  memsize = (long int)n1*n2*11;

  int tempgpuid;
  cudaGetDevice(&tempgpuid);
//  fprintf(stderr, " GPU %d memory allocated %f MB (%f GB) \n", tempgpuid, memsize*4./1024/1024, memsize*4./1024/1024/1024);

  cudaMemcpyToSymbol(p1_gpu,  &p1, sizeof(float*));
  cudaMemcpyToSymbol(p1x_gpu, &p1x, sizeof(float*));
  cudaMemcpyToSymbol(p1z_gpu, &p1z, sizeof(float*));
}

extern "C" void gpu_grad(float *p1_cpu, float *p1x_cpu, float *p1z_cpu)
{
  int n1, n2, nt;
  float *p1, *p1x, *p1z;

  cudaMemcpyFromSymbol(&n1,  n1_gpu, sizeof(int));
  cudaMemcpyFromSymbol(&n2,  n2_gpu, sizeof(int));
  cudaMemcpyFromSymbol(&nt,  nt_gpu, sizeof(int));
  cudaMemcpyFromSymbol(&p1,  p1_gpu, sizeof(float*));
  cudaMemcpyFromSymbol(&p1x, p1x_gpu, sizeof(float*));
  cudaMemcpyFromSymbol(&p1z, p1z_gpu, sizeof(float*));

  int nblocksx=ceil(1.*n1/BLOCK_SIZE);
  int nblocksy=ceil(1.*n2/BLOCK_SIZE);
  dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);
  dim3 dimGrid(nblocksx,nblocksy);


  cudaMemcpy(p1,  p1_cpu,  n1*n2*sizeof(float), cudaMemcpyHostToDevice);
  cudaMemcpy(p1x, p1x_cpu, n1*n2*sizeof(float), cudaMemcpyHostToDevice);
  cudaMemcpy(p1z, p1z_cpu, n1*n2*sizeof(float), cudaMemcpyHostToDevice);

  for(int it=0; it<nt; it++)
  {
    kernel_exec(grad_for_kernel<<<dimGrid, dimBlock>>>(p1, p1x, p1z));
    cudaMemcpy(&p1x_cpu[(long int)n1*n2*it], p1x, n1*n2*sizeof(float), cudaMemcpyDeviceToHost);
    cudaMemcpy(&p1z_cpu[(long int)n1*n2*it], p1z, n1*n2*sizeof(float), cudaMemcpyDeviceToHost);
  }
}


