program hess_psf
!*** Constructs the Gauss-Newton Hessian using point-spread functions (PSF) ***
  use sep
  use mem_mod
  use fbi_mod
!  use fbi_energ_mod
!  use tconv_mod
!  use solver_mod
!  use solver_wjreg_mod
!  use cgstep_mod
  
  implicit none

  integer                             :: stat  
  integer                             :: nt, n1, n2, wfld_j, i, gpu_n, gpu_o
  integer                             :: np1, np2, dp1, dp2
  integer                             :: ip1, ip2
  integer                             :: sou_nx, sou_ox, sou_dx, sou_z, sou_x, isou
  integer                             :: rec_nx, rec_ox, rec_dx, rec_z, rec_x, irec
  integer                             :: bc1, bc2, bc3, bc4, ntau, otau
  real                                :: dt, d1, o1, d2, o2
  real                                :: lag1, lag2

  logical                             :: dipole

  real, dimension(:), allocatable     :: wavelet
  real, dimension(:,:), allocatable   :: vel
  real, dimension(:,:,:), allocatable :: data_obs, refl, temp1, model

  real, dimension(:), allocatable     :: data, bc

  integer, dimension(8)               :: timer0, timer1
  real                                :: timer_sum=0.

  CALL sep_init()

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer0)
!---------------------------------------------------------------------------

  call from_history("n1", n1)
  call from_history("o1", o1)
  call from_history("d1", d1)
  call from_history("n2", n2)
  call from_history("o2", o2)
  call from_history("d2", d2)

  call from_param("ntau", ntau, 1)
  call from_param("otau", otau, 0)
  if(ntau<1-otau) write(0,*) "WARNING: zero subsurface axis is not included"

  call from_param("nt", nt)
  call from_aux("wavelet", "d1", dt)

  call from_param("bc1", bc1)
  call from_param("bc2", bc2)
  call from_param("bc3", bc3)
  call from_param("bc4", bc4)
  call from_param("wfld_j", wfld_j)

  call from_param("sou_nx", sou_nx)
  call from_param("sou_ox", sou_ox)
  call from_param("sou_dx", sou_dx)
  call from_param("sou_z", sou_z)
  if(sou_ox < 1 .or. (sou_nx-1)*sou_dx+sou_ox > n2) call erexit("sou axis values are not correct")
  if(sou_z < 1 .or. sou_z > n1) call erexit("sou_z value is not correct")

  call from_param("rec_nx", rec_nx)
  call from_param("rec_ox", rec_ox)
  call from_param("rec_dx", rec_dx)
  call from_param("rec_z", rec_z)
  if(rec_ox < 1 .or. (rec_nx-1)*rec_dx+rec_ox > n2) call erexit("rec axis values are not correct")
  if(rec_z < 1 .or. rec_z > n1) call erexit("rec_z value is not correct")

  call from_param("dipole", dipole, .false.)

  call from_param("gpu_n", gpu_n)
  call from_param("gpu_o", gpu_o)

  call from_param("lag1", lag1)
  call from_param("lag2", lag2)
  call from_param("dp1", dp1)
  call from_param("dp2", dp2)

  np1 = n1/dp1
  np2 = n2/dp2

  call init1(n1, o1, d1, n2, o2, d2, bc1, bc2, bc3, bc4)
  call init2(sou_nx, sou_ox, sou_dx, sou_z, rec_nx, rec_ox, rec_dx, rec_z)
  call init3(nt, dt, wfld_j, ntau, otau, gpu_n, gpu_o)
  call init4(dipole)

  call to_history("n1", n1)
  call to_history("o1", o1)
  call to_history("d1", d1)
  call to_history("n2", n2)
  call to_history("o2", o2)
  call to_history("d2", d2)
  call to_history("n3", ntau)
  call to_history("o3", otau)
  call to_history("d3", 1.)

  call mem_alloc1d(wavelet, nt)
  call mem_alloc2d(vel, n1, n2)
  call mem_alloc3d(model, n1, n2, ntau)

  call sep_read(vel)
  call sep_read(wavelet, "wavelet")

!  call lowcut_init(n1,n2,0.85)
  
  call init_alloc(wavelet)
  call model_update(vel)

  call mem_alloc3d(data_obs, nt, rec_nx, sou_nx)
  call mem_alloc1d(data, n1*n2*ntau)
  call mem_alloc3d(refl, n1, n2, ntau)
  call mem_alloc3d(temp1, n1, n2, ntau)
  write(0,*) "Using dipole source? ",dipole
  write(0,*) " "
  write(0,*) "Reading data..."

  write(0,*) "Hessian construction running!"
!  do ip2 = 1+lag2,np2-lag2,dp2
!    write(0,*) "ip2=",ip2,"np2=",np2
!    do ip1 = 1+lag1,np1-lag1,dp1
!      write(0,*) "ip1=",ip1,"np1=",np1
!      refl = 0.
!      call L_lop2(.false., .false., refl, data_obs)
!      call L_lop2(.true.,  .false., temp1, data_obs)
!      model(-lag1/2+ip1:ip1+lag1/2,-lag2/2+ip2:ip2+lag2/2,1) = &
!      temp1(-lag1/2+ip1:ip1+lag1/2,-lag2/2+ip2:ip2+lag2/2,1)
!    end do
!  end do

  write(0,*) "Temporal: Hessian construction in one point!"
  ip1=n1/2
  ip2=n2/2
  refl = 0.
!   do i=1,sou_nx
!     call sep_read(data_obs(:,:,i), "data_obs")
!   end do
  refl(ip1,ip2,1) = 1. 
  write(0,*) "Modeling part..."
  call L_lop2(.false., .false., refl, data_obs)
write(0,*) "Norm product data, ", norm2(data_obs)
  write(0,*) "Migration part..."
  call L_lop2(.true.,  .false., temp1, data_obs)
  model(-lag1/2+ip1:ip1+lag1/2,-lag2/2+ip2:ip2+lag2/2,1) = & 
  temp1(-lag1/2+ip1:ip1+lag1/2,-lag2/2+ip2:ip2+lag2/2,1)

  write(0,*) "Finishing Hessian construction, now writing!"
  call sep_write(temp1)

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  write(0,*) "DONE!"
!---------------------------------------------------------------------------

END PROGRAM
