module hess_wemva_mod

  use sep
  use fbi_mod
!  use fbi_energ_mod
  use H_mod
 
  implicit none

  real, private                            :: lambda, signo=+1.
  real, dimension(:), allocatable, private :: model1, model2, data1, data2

  logical, private                         :: wemva
  contains

  subroutine hess_wemva_init(nd, nm, lambda_in, wemva_in)
    integer    :: nd, nm
    real       :: lambda_in
    logical    :: wemva_in
    lambda = lambda_in
    wemva  = wemva_in
    allocate(model1(nm/2)) 
    allocate(model2(nm/2)) 
    allocate(data1(nd/2)) 
    allocate(data2(nd/2)) 
  end subroutine

  function hess_wemva(adj, add, model, data) result(stat)
    integer                      :: stat 
    logical, intent(in)          :: adj, add 
    real, dimension(:)           :: model, data
    call hess_wemva2(adj, add, model, data)
    stat=0
  end function

  subroutine hess_wemva2(adj, add, model, data)
    logical, intent(in)          :: adj, add
    real, dimension(:)           :: model
    real, dimension(:)           :: data
    if(adj) then
      if(.not. add) model = 0.
      data1 = data(1:size(data)/2)
      data2 = data(size(data)/2+1:size(data))
      data2 = signo*data1+lambda*data2

!      model1 = Hgn*data1
      call H_lop2(.true., .false., model1, data1)
      if(wemva) then 
        call W_lop2(.true., .false., model2, data2)
        model2=-1.0*model2 !Uncomment after experiment maximizing ||Imig-HwB||^2
      else
        model2=0
      end if

      model(1:size(model)/2)             = model(1:size(model)/2)            +model1
      model(size(model)/2+1:size(model)) = model(size(model)/2+1:size(model))+model2

    else
      if(.not. add) data = 0.
      model1=model(1:size(model)/2)
      model2=model(size(model)/2+1:size(model))

!      data1 = Hgn*model1
      call H_lop2(.false., .false., model1, data1)
      if(wemva) then 
        call W_lop2(.false., .false., model2, data2)
        data2=-1.0*data2!Uncomment after experiment maximizing ||Imig-HwB||^2
      else
        data2=0
      end if
      data(1:size(data)/2)            = data(1:size(data)/2)           +data1+signo*data2
      data(size(data)/2+1:size(data)) = data(size(data)/2+1:size(data))+lambda*data2
    end if

  end subroutine

end module
