PROGRAM lsq
!*** Implements the least-squares solvers in modules sdstep_mod and cgstep_mod ***
  use sep
  use mem_mod
  use fbi_mod
!  use fbi_energ_mod
!  use tconv_mod
  use solver_mod
  use solver_smp1_mod
!  use solver_wjreg_mod
  use cgstep_mod
  use lowcut_mod
  
  implicit none
  
  integer                             :: nt, n1, n2, wfld_j, i, gpu_n, gpu_o, niter, nreg
  integer                             :: sou_nx, sou_ox, sou_dx, sou_z, sou_x, isou
  integer                             :: rec_nx, rec_ox, rec_dx, rec_z, rec_x, irec
  integer                             :: bc1, bc2, bc3, bc4, ntau, otau
  real                                :: dt, d1, o1, d2, o2

  real                                :: alpha, eps=0

  logical                             :: dipole

  real, dimension(:), allocatable     :: wavelet
  real, dimension(:,:), allocatable   :: vel!, deltaB
  real, dimension(:,:,:), allocatable :: data_obs, Imig

  real, dimension(:), allocatable     :: data, model, bc

  real, dimension(:,:), allocatable   :: rmov, mmov

  integer, dimension(8)               :: timer0, timer1
  real                                :: timer_sum=0.

  logical                             :: reg, precond
  
  CALL sep_init()

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer0)
!---------------------------------------------------------------------------

  call from_history("n1", n1)
  call from_history("o1", o1)
  call from_history("d1", d1)
  call from_history("n2", n2)
  call from_history("o2", o2)
  call from_history("d2", d2)

  call from_param("ntau", ntau, 1)
  call from_param("otau", otau, 0)
  if(ntau<1-otau) write(0,*) "WARNING: zero subsurface axis is not included"

  call from_param("nt", nt)
  call from_aux("wavelet", "d1", dt)


  call from_param("eps", eps, 0.0)

  call from_param("bc1", bc1)
  call from_param("bc2", bc2)
  call from_param("bc3", bc3)
  call from_param("bc4", bc4)
  call from_param("wfld_j", wfld_j)

  call from_param("sou_nx", sou_nx)
  call from_param("sou_ox", sou_ox)
  call from_param("sou_dx", sou_dx)
  call from_param("sou_z", sou_z)
  if(sou_ox < 1 .or. (sou_nx-1)*sou_dx+sou_ox > n2) call erexit("sou axis values are not correct")
  if(sou_z < 1 .or. sou_z > n1) call erexit("sou_z value is not correct")

  call from_param("rec_nx", rec_nx)
  call from_param("rec_ox", rec_ox)
  call from_param("rec_dx", rec_dx)
  call from_param("rec_z", rec_z)
  if(rec_ox < 1 .or. (rec_nx-1)*rec_dx+rec_ox > n2) call erexit("rec axis values are not correct")
  if(rec_z < 1 .or. rec_z > n1) call erexit("rec_z value is not correct")

  call from_param("dipole", dipole, .false.)

  call from_param("gpu_n", gpu_n)
  call from_param("gpu_o", gpu_o)

  call from_param("niter", niter)
  call from_param("alpha", alpha, 0.)

  call init1(n1, o1, d1, n2, o2, d2, bc1, bc2, bc3, bc4)
  call init2(sou_nx, sou_ox, sou_dx, sou_z, rec_nx, rec_ox, rec_dx, rec_z)
  call init3(nt, dt, wfld_j, ntau, otau, gpu_n, gpu_o)
  call init4(dipole)
!  call init4(dipole, .false., alpha)

  call to_history("n1", n1)
  call to_history("o1", o1)
  call to_history("d1", d1)
  call to_history("n2", n2)
  call to_history("o2", o2)
  call to_history("d2", d2)
  call to_history("n3", ntau)
  call to_history("o3", otau)
  call to_history("d3", 1.)

  call to_history("n1", n1,    "model_mov")
  call to_history("o1", o1,    "model_mov")
  call to_history("d1", d1,    "model_mov")
  call to_history("n2", n2,    "model_mov")
  call to_history("o2", o2,    "model_mov")
  call to_history("d2", d2,    "model_mov")
  call to_history("n3", niter, "model_mov")
  call to_history("o3", 1,     "model_mov")
  call to_history("d3", 1,     "model_mov")

  call to_history("n1", n1,    "resid_mov")
  call to_history("o1", o1,    "resid_mov")
  call to_history("d1", d1,    "resid_mov")
  call to_history("n2", n2,    "resid_mov")
  call to_history("o2", o2,    "resid_mov")
  call to_history("d2", d2,    "resid_mov")
  call to_history("n3", niter, "resid_mov")
  call to_history("o3", 1,     "resid_mov")
  call to_history("d3", 1,     "resid_mov")

  call mem_alloc1d(wavelet, nt)
  call mem_alloc2d(vel, n1, n2)

  call sep_read(vel)
  call sep_read(wavelet, "wavelet")

  if(eps /= 0) then
    call from_aux("filter", "n1", nreg)
    ALLOCATE(bc(nreg))
    call sep_read(bc, "filter")
!    call tconv1_init1(n1*n2,nreg)
!    call tconv1_init2(bc)
  end if
  
!  call lowcut_init(n1,n2,0.85)
  
  call init_alloc(wavelet)
  call model_update(vel)

  call mem_alloc3d(data_obs, nt, rec_nx, sou_nx)
  call mem_alloc3d(Imig, n1, n2, ntau)
  call mem_alloc1d(data, n1*n2*ntau)
  call mem_alloc1d(model, n1*n2)

  call mem_alloc2d(mmov, n1*n2,niter)
  call mem_alloc2d(rmov, n1*n2*ntau,niter)

  write(0,*) "Using dipole source? ",dipole
  write(0,*) " "
  write(0,*) "Reading data..."
  do i=1,sou_nx
    call sep_read(data_obs(:,:,i), "data")
  end do
  call data_update(data_obs)

  write(0,*) "WEMVA inversion running!"
  do i=1,ntau
    call sep_read(Imig(:,:,i), "Imig")
  end do
  data = 1.0*reshape(Imig,(/n1*n2*ntau/))
!  if(eps==0) then
!    call solver_tiny(model,data,W_lop,cgstep,niter)
    call solver_smp(model,data,W_lop,cgstep,niter,mmov=mmov,rmov=rmov,verb=.true.)
!  else
!    call solver_wjreg(model,data,W_lop,tconv1_lop,cgstep,nreg,niter,eps) !,lowcut_op)
!  end if
  write(0,*) "Finishing WEMVA inversion, now writing!"
  call sep_write(model)

  do i=1,niter
    call sep_write(mmov(:,i), "model_mov")
    call sep_write(rmov(:,i), "resid_mov")
  end do  

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  write(0,*) "DONE!"
!---------------------------------------------------------------------------

END PROGRAM
