PROGRAM lwivu
!*** Implements the least-squares solvers in modules sdstep_mod and cgstep_mod ***
  use sep
  use mem_mod
  use fbi_mod
!  use fbi_energ_mod
  use tconv_mod
  use solver_smp1_mod
  use solver_reg1_mod
  use cgstep_mod
!  use lowcut_mod
!  use grad1_mod
  use adj_mod
  use H_mod
  use hess_wemva_mod
  
  implicit none
  
  integer                             :: nt, n1, n2, nlag1, nlag2, wfld_j, i, gpu_n, gpu_o
  integer                             :: niter, nreg, nd, nm
  integer                             :: sou_nx, sou_ox, sou_dx, sou_z, sou_x, isou
  integer                             :: rec_nx, rec_ox, rec_dx, rec_z, rec_x, irec
  integer                             :: bc1, bc2, bc3, bc4, ntau, otau
  real                                :: dt, d1, o1, d2, o2
  real                                :: alpha, lambda, eps

  real, dimension(:), allocatable     :: wavelet, r1, r2
  real, dimension(:,:), allocatable   :: vel!, deltaB
  real, dimension(:,:,:), allocatable :: data_obs, Imig
  real, dimension(:), allocatable     :: data, model, bc, m0
  real, dimension(:,:), allocatable   :: rmov, mmov

  integer, dimension(8)               :: timer0, timer1
  real                                :: timer_sum=0.

  logical                             :: dipole
  logical                             :: reg, precond, wemva, verb, priori

  CALL sep_init()

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer0)
!---------------------------------------------------------------------------

  call from_history("n1", n1)
  call from_history("o1", o1)
  call from_history("d1", d1)
  call from_history("n2", n2)
  call from_history("o2", o2)
  call from_history("d2", d2)

  call from_aux("hess", "n3", i)
  nlag1=(i-1)/2
  call from_aux("hess", "n4", i)
  nlag2=(i-1)/2

  call from_param("ntau", ntau, 1)
  call from_param("otau", otau, 0)
  if(ntau<1-otau) write(0,*) "WARNING: zero subsurface axis is not included"

  call from_param("nt", nt)
  call from_aux("wavelet", "d1", dt)


!  call from_param("reg", reg, .false.)

  call from_param("bc1", bc1)
  call from_param("bc2", bc2)
  call from_param("bc3", bc3)
  call from_param("bc4", bc4)
  call from_param("wfld_j", wfld_j)

  call from_param("sou_nx", sou_nx)
  call from_param("sou_ox", sou_ox)
  call from_param("sou_dx", sou_dx)
  call from_param("sou_z", sou_z)
  if(sou_ox < 1 .or. (sou_nx-1)*sou_dx+sou_ox > n2) call erexit("sou axis values are not correct")
  if(sou_z < 1 .or. sou_z > n1) call erexit("sou_z value is not correct")

  call from_param("rec_nx", rec_nx)
  call from_param("rec_ox", rec_ox)
  call from_param("rec_dx", rec_dx)
  call from_param("rec_z", rec_z)
  if(rec_ox < 1 .or. (rec_nx-1)*rec_dx+rec_ox > n2) call erexit("rec axis values are not correct")
  if(rec_z < 1 .or. rec_z > n1) call erexit("rec_z value is not correct")

  call from_param("dipole", dipole, .false.)

  call from_param("gpu_n", gpu_n)
  call from_param("gpu_o", gpu_o)

  call from_param("niter", niter)
  call from_param("lambda", lambda,0.1)
  call from_param("wemva", wemva, .true.)
  call from_param("verbose", verb, .true.)
  call from_param("priori", priori, .false.)
  call from_param("alpha", alpha, 0.)

  call from_param("eps", eps, 0.)

!  if(reg) then
!    wemva = .true.
!  end if

  call init1(n1, o1, d1, n2, o2, d2, bc1, bc2, bc3, bc4)
  call init2(sou_nx, sou_ox, sou_dx, sou_z, rec_nx, rec_ox, rec_dx, rec_z)
  call init3(nt, dt, wfld_j, ntau, otau, gpu_n, gpu_o)
  call init4(dipole)
!  call init4(dipole, .false., alpha)

  call to_history("n1", n1)
  call to_history("o1", o1)
  call to_history("d1", d1)
  call to_history("n2", n2)
  call to_history("o2", o2)
  call to_history("d2", d2)
  call to_history("n3", ntau)
  call to_history("o3", otau)
  call to_history("d3", 1.)

  call to_history("n1", n1, "deltaB")
  call to_history("o1", o1, "deltaB")
  call to_history("d1", d1, "deltaB")
  call to_history("n2", n2, "deltaB")
  call to_history("o2", o2, "deltaB")
  call to_history("d2", d2, "deltaB")
  call to_history("n3", ntau, "deltaB")
  call to_history("o3", otau, "deltaB")
  call to_history("d3", 1., "deltaB")

  call to_history("n1", n1,    "modeldr_mov")
  call to_history("o1", o1,    "modeldr_mov")
  call to_history("d1", d1,    "modeldr_mov")
  call to_history("n2", n2,    "modeldr_mov")
  call to_history("o2", o2,    "modeldr_mov")
  call to_history("d2", d2,    "modeldr_mov")
  call to_history("n3", niter, "modeldr_mov")
  call to_history("o3", 1,     "modeldr_mov")
  call to_history("d3", 1,     "modeldr_mov")

  if(wemva) then
    call to_history("n1", n1,    "modeldb_mov")
    call to_history("o1", o1,    "modeldb_mov")
    call to_history("d1", d1,    "modeldb_mov")
    call to_history("n2", n2,    "modeldb_mov")
    call to_history("o2", o2,    "modeldb_mov")
    call to_history("d2", d2,    "modeldb_mov")
    call to_history("n3", niter, "modeldb_mov")
    call to_history("o3", 1,     "modeldb_mov")
    call to_history("d3", 1,     "modeldb_mov")
  end if

  call to_history("n1", n1,    "residdr1_mov")
  call to_history("o1", o1,    "residdr1_mov")
  call to_history("d1", d1,    "residdr1_mov")
  call to_history("n2", n2,    "residdr1_mov")
  call to_history("o2", o2,    "residdr1_mov")
  call to_history("d2", d2,    "residdr1_mov")
  call to_history("n3", niter, "residdr1_mov")
  call to_history("o3", 1,     "residdr1_mov")
  call to_history("d3", 1,     "residdr1_mov")

  call to_history("n1", niter, "r1")
  call to_history("o1", 1,     "r1")
  call to_history("d1", 1,     "r1")

  if(wemva) then
    call to_history("n1", n1,    "residdr2_mov")
    call to_history("o1", o1,    "residdr2_mov")
    call to_history("d1", d1,    "residdr2_mov")
    call to_history("n2", n2,    "residdr2_mov")
    call to_history("o2", o2,    "residdr2_mov")
    call to_history("d2", d2,    "residdr2_mov")
    call to_history("n3", niter, "residdr2_mov")
    call to_history("o3", 1,     "residdr2_mov")
    call to_history("d3", 1,     "residdr2_mov")

    call to_history("n1", niter, "r2")
    call to_history("o1", 1,     "r2")
    call to_history("d1", 1,     "r2")
  end if

  call mem_alloc1d(wavelet, nt)
  call mem_alloc2d(vel, n1, n2)

  call sep_read(vel)
  call sep_read(wavelet, "wavelet")

  
  call init_alloc(wavelet)
  call model_update(vel)

  nd = 2*n1*n2*ntau
  nm = 2*n1*n2

  if(eps /= 0) then
  if(.not. wemva) then
    call from_aux("filter", "n1", nreg)
    ALLOCATE(bc(nreg))
    call sep_read(bc, "filter")
    call tconv1_init1(nm,nreg)
    call tconv1_init2(bc)
    write(0,*) 'bc',bc
  end if
  end if
 
!  call lowcut_init(n1,n2,0.85)

!  if(reg) then
!    call hess_wemva_init(nd, nm, 0.0, hess, wemva) 
!  else
    call H_init(n1, n2, nlag1, nlag2)
    call hess_wemva_init(nd, nm, lambda, wemva) 
!  end if

  call mem_alloc3d(data_obs, nt, rec_nx, sou_nx)
  call mem_alloc3d(Imig, n1, n2, ntau)
  call mem_alloc1d(data, nd)
  call mem_alloc1d(model, nm)

  if(priori) then
    call mem_alloc1d(m0, nd)
    call sep_read(m0(1:nd/2),"m0")
  end if 

  call mem_alloc2d(mmov, nm,niter)
  call mem_alloc2d(rmov, nd,niter)

  call mem_alloc1d(r1, niter)
  if(wemva) call mem_alloc1d(r2, niter)

  write(0,*) "Using dipole source? ",dipole
  write(0,*) " "
  write(0,*) "Reading data..."
  do i=1,sou_nx
    call sep_read(data_obs(:,:,i), "data")
  end do
  call data_update(data_obs)

  write(0,*) "LWIVU running!"
  do i=1,ntau
    call sep_read(Imig(:,:,i), "Imig")
  end do
  data(1:nd/2)    = reshape(Imig,(/n1*n2*ntau/))
  data(nd/2+1:nd) = -lambda*reshape(Imig,(/n1*n2*ntau/))
!  if(priori) then
!    m0=0
!    m0(1:nd/2)    = reshape(Imig,(/n1*n2*ntau/))
!    m0(nd/2+1:nd) =( 1./reshape(vel,(/n1*n2/)))**2
!  endif
  if(.not. wemva .and. eps /= 0.) then
    write(0,*) 'Regularized CD, eps=',eps  
    call solver_reg1(model,data,hess_wemva,tconv1_lop,cgstep,nd,niter,eps=eps,mmov=mmov,rmov=rmov,verb=verb)
  else
    if(priori) then
      call solver_smp(model,data,hess_wemva,cgstep,niter,m0=m0,mmov=mmov,rmov=rmov,verb=verb)
    else
      call solver_smp(model,data,hess_wemva,cgstep,niter,mmov=mmov,rmov=rmov,verb=verb)
    end if
  end if
  write(0,*) "Finishing LWIVU, now writing!"
  call sep_write(model(1:nm/2))
  call sep_write(model(nm/2+1:nm),"deltaB")
  do i=1,niter
    call sep_write(mmov(1     :nm/2,i),"modeldr_mov")
    call sep_write(rmov(1     :nd/2,i),"residdr1_mov")
    r1(i)=sqrt(dot_product(rmov(1:nd/2,i),rmov(1:nd/2,i)))
    if(wemva) then
      call sep_write(mmov(nm/2+1:nm,i),  "modeldb_mov")
      call sep_write(rmov(nd/2+1:nd,i),  "residdr2_mov")
      r2(i)=sqrt(dot_product(rmov(nd/2+1:nd,i),rmov(nd/2+1:nd,i)))
    end if
  end do  
  call sep_write(r1,"r1")
  if(wemva) call sep_write(r2,"r2")

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  write(0,*) "DONE!"
!---------------------------------------------------------------------------

END PROGRAM
